﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security;
using System.Security.Permissions;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Media;
using System.Windows.Threading;
using ImageNexus.BenScharbach.Common.AttributesCodeInjector.AttsException;
using ImageNexus.BenScharbach.Common.AttributesCodeInjector.AttsProgressBar;
using ImageNexus.BenScharbach.Common.AttributesCodeInjector.AttsThreading;
using ImageNexus.BenScharbach.Common.AttsInjectorHelper;
using ImageNexus.BenScharbach.Common.Extensions;
using ImageNexus.BenScharbach.Common.FileProcessor;
using ImageNexus.BenScharbach.Common.MEFLoggerInterfaces.Enums;

namespace ImageNexus.BenScharbach.Utilities.EverNoteDropper
{
    // 1/13/2013
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        // tracker folder, stored in the ProgramData section.
        private const string EverNoteTrackerFolderName = "TrackerData";
        // source folder
        private const string EverNoteSourceFolderNameDefault = "EverNote Source";
        // dropper folder
        private const string EverNoteDropperFolderNameDefault = "EverNote DropBox";
        // tracker name
        private const string EverNoteTrackerFileName = "TrackerDictionary.xml";
        // programData internal path
        private const string EverNoteProgramDataPathName = @"Image-Nexus\EverNoteDropper";

        private static string _everNoteTrackerPath;
        private static string _everNoteDropPath;
        private static string _everNoteSourcePath;
        private static string _dropperLogPath;

        // MyDocs location on this computer
        private static string _myDocPath;
        // Dictionary tracker, serialized to the tracker folder.
        private static readonly Dictionary<string, string> EverNoteTrackerDictionary = new Dictionary<string, string>();

        // 1/14/2013
        private static RsltWindow _resultsWindow;

        #region Properties

        // 1/13/2013
        /// <summary>
        /// Gets the EverNote source folder name.
        /// </summary>
        public string EverNoteSourceFolderName
        {
            [GuiThread(DispatcherPriority.Normal)]
            get { return txtEverNoteSourceFolder.Text ?? EverNoteSourceFolderNameDefault; }
        }

        // 1/13/2013
        /// <summary>
        /// Gets the EverNote dropper folder name.
        /// </summary>
        public string EverNoteDropperFolderName
        {
            [GuiThread(DispatcherPriority.Normal)]
            get { return txtEverNoteDropFolder.Text ?? EverNoteDropperFolderNameDefault; }
        }

        #endregion

        /// <summary>
        /// Constructor
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            // Create 'Results' window
            _resultsWindow = new RsltWindow();

            // 1/13/2013 - subscribe to the OnCatch attributes 'UserFriendly' events
            OnCatchAllExceptionHandlingAttribute.UserFriendlyErrorMessageOccurs += OnCatchAllExceptionHandlingAttribute_UserFriendlyMessageOccurs;

            _myDocPath = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            var prmDataPath = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
            _dropperLogPath = string.Concat(prmDataPath, @"\", EverNoteProgramDataPathName, @"\");

            // create tracker path
            _everNoteTrackerPath = String.Concat(prmDataPath, @"\", EverNoteProgramDataPathName, @"\", EverNoteTrackerFolderName, @"\");
            // create destination path
            _everNoteDropPath = String.Concat(_myDocPath, @"\", EverNoteDropperFolderName, @"\");
            // create source path
            _everNoteSourcePath = String.Concat(_myDocPath, @"\", EverNoteSourceFolderName, @"\");

            // check if source folder exist; if not, create it and exit app.
            if (!Directory.Exists(_everNoteSourcePath))
            {
                Directory.CreateDirectory(_everNoteSourcePath);
                return;
            }

            // check if source folder is empty
            if (!Directory.GetDirectories(_everNoteSourcePath).Any() && !Directory.GetFiles(_everNoteSourcePath).Any())
            {
                // let user know it is complete
                OnCatchAllExceptionHandlingAttribute.SetUserFriendlyErrorMessage(null, "The EverNote source folder is empty.");
            }

            // check if tracker folder exist; if not, create it.
            if (!Directory.Exists(_everNoteTrackerPath))
                Directory.CreateDirectory(_everNoteTrackerPath);

            // check if destination folder exist; if not, create it.
            if (!Directory.Exists(_everNoteDropPath))
                Directory.CreateDirectory(_everNoteDropPath);

            // Load prior 'Tracker' data
            LoadTrackerDictionary();
        }

        // 1/13/2013
        /// <summary>
        /// Handles the User-Friendly error messages thrown by the <see cref="OnCatchAllExceptionHandlingAttribute"/>.
        /// </summary>
        [GuiThread(DispatcherPriority.Normal)]
        private void OnCatchAllExceptionHandlingAttribute_UserFriendlyMessageOccurs(object sender, OnCatchAllExceptionHandlingAttribute.UserFriendlyErrorMessageEventArgs e)
        {
            lblMessage.Content = e.Message;
            lblMessage.Foreground = e.Foreground;
        }

        /// <summary>
        /// Handles the Sync button.
        /// </summary>
        private void btnEverNoteDropperSync_Click(object sender, RoutedEventArgs e)
        {
            // Hide window
            _resultsWindow.Hide();
            _resultsWindow.ClearAll();

            // does the EverNote sync operation
            if (!DoEverNoteCompare(null, this))
                return;

            // Show Results Window
            _resultsWindow.Show();

            // does the XML save operation
            SaveTrackerData();  
        }

        /// <summary>
        /// Compares the EverNote source folder with the drop folder, and copies where necessary using the
        /// tracker dictionary data.
        /// </summary>
        [ProgressBarWindow]
        private static bool DoEverNoteCompare(ProgressBarWindowAttribute.ProgressBarItem pbWindow, MainWindow mainWindow)
        {
            if (mainWindow == null)
                throw new ArgumentNullException("mainWindow");

            // Create DateTime string for Logger
            var dateTimeString = DateTime.Now.Month + "_" + DateTime.Now.Day + "_" + DateTime.Now.Year + "_" +
                                 DateTime.Now.Minute + "_" +
                                 DateTime.Now.Second;

            // Start Logger
            var fileCount = Processor.GetFileCount(_everNoteSourcePath);
            var logName = "EverNoteCompareOperation_" + dateTimeString + ".txt";
            OnMefLoggerAttribute.Logger.SetLoggerOutputPath(LoggerTypeEnum.StreamWriterLog, EverNoteProgramDataPathName);
            OnMefLoggerAttribute.Logger.SetNumberOfFilesToProcess(LoggerTypeEnum.StreamWriterLog, fileCount);
            OnMefLoggerAttribute.Logger.StartStreamWriterLogfile(logName, "Compare Files...");
           
            // copy and compare EverNote directories
            if (!EverNoteCompare(pbWindow, mainWindow, _everNoteSourcePath))
                return false;

            // Stop Logger
            OnMefLoggerAttribute.Logger.EndStreamWriterLogFile();

            // let user know it is complete
            OnCatchAllExceptionHandlingAttribute.SetUserFriendlyMessage(null, "The Sync operation is complete.");

            return true;
        }

        /// <summary>
        /// Helper method, which saves the Dictionary as XML serialized data.
        /// </summary>
        [OnCatchAllExceptionHandling(true, typeof(bool), false, "EverNote save XML operation failed.")]
        private static void SaveTrackerData()
        {
            // Serialize the XML dictionary.
            Processor.SaveDictionaryXmlSerializedStream<Dictionary<string, string>, string, string>(EverNoteTrackerDictionary,
                                                                                                    EverNoteTrackerFileName,
                                                                                                    _everNoteTrackerPath);
        }

        /// <summary>
        /// Recurisively compares all files in the source folder to the tracking folder, and
        /// copies files over where necessary.
        /// </summary>
        [OnCatchAllExceptionHandling(true, typeof(bool), false, "EverNote Sync operation failed.")]
        private static bool EverNoteCompare(ProgressBarWindowAttribute.ProgressBarItem pbWindow, MainWindow mainWindow, string directory)
        {
            if (mainWindow == null)
                throw new ArgumentNullException("mainWindow");

            // iterate all directories in this directory
            var enumerateDirectories = Directory.EnumerateDirectories(directory).ToArray();
            var count = enumerateDirectories.Length;
            for (var index = 0; index < count; index++)
            {
                var enumDirectory = enumerateDirectories[index];

                // set progress
                pbWindow.SetProgressBarHeaderText(enumDirectory, index, count);

                // next directory
                EverNoteCompare(pbWindow, mainWindow, enumDirectory);
            }

            IterateFiles(pbWindow, mainWindow, directory);
           
            return true;
        }

        /// <summary>
        /// Helper method, which iterates all files in the given <paramref name="directory"/>.
        /// </summary>
        private static void IterateFiles(ProgressBarWindowAttribute.ProgressBarItem pbWindow, MainWindow mainWindow, string directory)
        {
            if (mainWindow == null)
                throw new ArgumentNullException("mainWindow");

            var everNoteSourceFolderName = mainWindow.EverNoteSourceFolderName;
            var everNoteDropperFolderName = mainWindow.EverNoteDropperFolderName;

            // iterate all files in this directory
            var enumerateFiles = Directory.EnumerateFiles(directory).ToArray();
            var count = enumerateFiles.Length;
            for (var index = 0; index < count; index++)
            {
                var file = enumerateFiles[index];
                var destFile = file.Replace(everNoteSourceFolderName, everNoteDropperFolderName);

                // Log file to be compared into EventLog.
                OnMefLoggerAttribute.Logger.WriteItem(LoggerTypeEnum.StreamWriterLog, string.Format("File to compare -> {0}; ", file), false, null);

                // if not there, include in dictionary and copy file to destination
                if (!EverNoteTrackerDictionary.ContainsKey(file))
                {
                    // Note: When set to TRUE, then 'AddUpdate' operation is used.
                    EverNoteTrackerDictionary.AddRemoveItemFromCollection(file, file, true);

                    // Copy file over
                    CopyFile(file, destFile);

                    // Log file to be compared into EventLog.
                    OnMefLoggerAttribute.Logger.WriteItem(LoggerTypeEnum.StreamWriterLog, string.Format("File copied -> {0}; ", destFile), true, null);
                    _resultsWindow.AddResultsItem(string.Format("File copied -> {0}; ", destFile), true, Brushes.Green);
                }
                else
                {
                    _resultsWindow.AddResultsItem(string.Format("File skipped -> {0}; ", destFile), false, Brushes.DarkRed);
                }

                // set progress
                pbWindow.SetProgressBarHeaderText(file, index, count);
            }
        }

        /// <summary>
        /// Helper method, which copies the files from the source to dest.
        /// </summary>
        private static void CopyFile(string file, string destFile)
        {
            try
            {
                // give IO access
                GiveFileIOPermissions(destFile);

                // copy file to destination
                File.Copy(file, destFile, true);
            }
            catch (UnauthorizedAccessException ex)
            {
                _resultsWindow.AddResultsItem(string.Format("File Error -> {0}; ", destFile), false, Brushes.Red);
            }
            catch (DirectoryNotFoundException)
            {
                // create FileInfo wrapper
                var fileInfo = new FileInfo(destFile);
                // create directory
                Directory.CreateDirectory(fileInfo.DirectoryName);

                // copy file to destination
                File.Copy(file, destFile, true);
            }
        }

        /// <summary>
        /// Loads the XML Dictionary from the 'Tracker' directory.
        /// </summary>
        [DoNotUseLogger]
        [OnCatchAllExceptionHandling(false, typeof(bool), false, "Loading Tracker data failed.")]
        private static void LoadTrackerDictionary()
        {
            // load the XML dictionary
            Dictionary<string, string> dictionary;
            if (!Processor.LoadDictionaryXmlSerializedStream<Dictionary<string, string>, string, string>(
                EverNoteTrackerFileName, _everNoteTrackerPath, out dictionary))
            {
                // let user know it is complete
                OnCatchAllExceptionHandlingAttribute.SetUserFriendlyErrorMessage(null, "Loading Tracker data failed.");
                return;
            }

            // copy data to current dictionary.
            EverNoteTrackerDictionary.Clear();
            foreach (var keyValuePair in dictionary)
            {
                EverNoteTrackerDictionary.Add(keyValuePair.Key, keyValuePair.Value);
            }
        }

        /// <summary>
        /// Closes application
        /// </summary>
        private void EverNoteDropper_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            Task.Factory.StartNew(SavePropertySettings);

            // Shutdown this application
            Application.Current.ShutdownMode = ShutdownMode.OnExplicitShutdown;
            Application.Current.Shutdown();
        }

        // 1/13/2013
        /// <summary>
        /// Saves property settings.
        /// </summary>
        [OnCatchAllExceptionHandling(false, typeof(bool), false)]
        private static void SavePropertySettings()
        {
            Properties.Settings.Default.Save();
        }

        // 1/13/2013
        /// <summary>
        /// Clears out the tracking data.
        /// </summary>
        [OnCatchAllExceptionHandling(false, typeof(bool), false, "Clear Tracking Data Failed!")]
        private void miClearTrackingData_Click(object sender, RoutedEventArgs e)
        {
            var messageResult = MessageBox.Show("This will clear out the 'Tracking' data.  Do you want to continue?",
                                                "Tracking Data",
                                                MessageBoxButton.YesNo, MessageBoxImage.Question);

            if (messageResult != MessageBoxResult.Yes) return;

            // clear internal dictionary too.
            EverNoteTrackerDictionary.Clear();

            // give IO access
            GiveFileIOPermissions(_everNoteTrackerPath);
            Directory.Delete(_everNoteTrackerPath, true);

            // let user know it is complete
            OnCatchAllExceptionHandlingAttribute.SetUserFriendlyMessage(null, "The clear 'Tracking' data completed successfully.");
        }

        /// <summary>
        /// Gives File IO AllAccess to the given path.
        /// </summary>
        [OnCatchAllExceptionHandling(false, typeof(bool), false, "Give FileIO access failed.")]
        private static void GiveFileIOPermissions(string path)
        {
            if (string.IsNullOrEmpty(path))
                throw new ArgumentNullException(path);

            var f2 = new FileIOPermission(FileIOPermissionAccess.AllAccess, path);
            try
            {
                f2.Demand();
            }
            catch (SecurityException s)
            {
                Console.WriteLine(s.Message);
            }
        }
    }
}
